from pathlib import Path

import fasttext
from typing import Tuple

CURRENT_DIR = Path(__file__).absolute().parent


class LanguageClassifier:
    def __init__(self):
        pretrained_lang_model = str(CURRENT_DIR / "lid.176.ftz")
        self.model = fasttext.load_model(pretrained_lang_model)

    def __call__(self, text: str) -> str:
        # remove all symbols
        text = text[:512].replace("\n", "")
        predictions = self.model.predict(text, k=1)
        score = predictions[1][0]
        pred = predictions[0][0].replace("__label__", "")
        return pred



from datetime import datetime
from enum import Enum
from typing import List

import typer
from pathlib import Path

from loguru import logger
from rich.progress import track

from llm_labeling_ui.db_schema import DBManager, Conversation
from llm_labeling_ui.utils import interactive_view_conversations, parse_tag

app = typer.Typer(
    add_completion=False,
    pretty_exceptions_show_locals=False,
    short_help="Conversation operations, such as remove prefix, remove duplicate, etc.",
)


@app.command(help="Remove conversation which is prefix of another conversation")
def remove_prefix(
    db_path: Path = typer.Option(None, exists=True, dir_okay=False),
    run: bool = typer.Option(False, help="run the command"),
):
    db = DBManager(db_path)
    conversations = db.all_conversations()
    logger.info(f"Total conversations: {len(conversations)}")

    import pygtrie

    trie = pygtrie.CharTrie()

    prefix_conversation_to_remove = []
    for it in track(conversations, description="building trie"):
        trie[it.merged_text()] = True

    for it in track(conversations, description="checking prefix"):
        if trie.has_subtrie(it.merged_text()):
            # 完全相等的 text 不会有 subtrie
            prefix_conversation_to_remove.append(it)

    logger.info(f"Found {len(prefix_conversation_to_remove)} prefix conversation")

    if run:
        for it in track(prefix_conversation_to_remove, description="removing"):
            db.delete_conversation(it.id)
        db.vacuum()


@app.command(help="Remove duplicate conversation only keep one of them")
def remove_duplicate(
    db_path: Path = typer.Option(None, exists=True, dir_okay=False),
    run: bool = typer.Option(False, help="run the command"),
):
    db = DBManager(db_path)
    conversations = db.all_conversations()
    logger.info(f"Total conversations: {len(conversations)}")

    conversation_to_remove = []
    merged_conversations = set()
    for it in track(conversations, description="finding duplicate"):
        merged_text = it.merged_text()
        if merged_text in merged_conversations:
            conversation_to_remove.append(it)
        else:
            merged_conversations.add(merged_text)

    logger.info(f"Found {len(conversation_to_remove)} duplicate conversation")

    if run:
        for it in track(conversation_to_remove, description="removing duplicates"):
            db.delete_conversation(it.id)
        db.vacuum()


@app.command(help="View conversation contain certain strings")
def view(
    db_path: Path = typer.Option(..., exists=True, dir_okay=False),
    search: List[str] = typer.Option([""], help="string to search"),
):
    db = DBManager(db_path)
    conversations = db.all_conversations(search_term=search)
    logger.info(f"Total conversations: {len(conversations)}")
    interactive_view_conversations(db, conversations, max_messages=5)


@app.command(help="Delete conversation contain certain string / tags")
def delete(
    db_path: Path = typer.Option(..., exists=True, dir_okay=False),
    search: str = typer.Option("", help="string to search"),
    run: bool = typer.Option(False, help="run the command"),
    tag: str = typer.Option(
        "", help="tag to filter conversations. key1,value1,key2,value2..."
    ),
    role: str = typer.Option(
        "all", help="role to search. user, assistant, system, all"
    ),
):
    tags = parse_tag(tag)
    db = DBManager(db_path)
    conversations = db.all_conversations()
    logger.info(f"Total conversations: {len(conversations)}")
    assert role in ["user", "assistant", "system", "all"]

    conversation_to_remove = []
    for it in track(conversations, description="finding duplicate"):
        if not it.contain_tags(tags):
            continue
        merged_text = it.merged_text(role=role)
        if search in merged_text:
            conversation_to_remove.append(it)

    logger.info(f"Found {len(conversation_to_remove)} conversations to remove")

    if run:
        for it in track(conversation_to_remove, description="removing conversations"):
            db.delete_conversation(it.id)
        db.vacuum()
    else:
        if len(conversation_to_remove) > 0:
            interactive_view_conversations(db, conversation_to_remove, max_messages=5)


@app.command(help="Delete string in conversation")
def delete_string(
    db_path: Path = typer.Option(None, exists=True, dir_okay=False),
    string: str = typer.Option(None, help="string to delete"),
    run: bool = typer.Option(False, help="run the command"),
):
    db = DBManager(db_path)
    conversations = db.all_conversations(search_term=string)
    logger.info(
        f"Total conversations {db.count_conversations()}, contains [{string}]: {len(conversations)}"
    )

    if run:
        for it in track(conversations, description="delete string"):
            it.data["prompt"] = it.data["prompt"].replace(string, "")
            for m in it.data["messages"]:
                m["content"] = m["content"].replace(string, "")
            it.updated_at = datetime.utcnow()
            db.update_conversation(it)
        db.vacuum()
    else:
        interactive_view_conversations(db, conversations, max_messages=5)


@app.command(help="Replace string in conversation")
def replace_string(
    db_path: Path = typer.Option(None, exists=True, dir_okay=False),
    search: str = typer.Option(..., help="string to search"),
    replace: str = typer.Option(..., help="replacement string"),
    run: bool = typer.Option(False, help="run the command"),
):
    db = DBManager(db_path)
    conversations = db.all_conversations()
    logger.info("Preview first 5 conversations:")
    max_preview = 5
    preview_count = 0

    matched_conversations = []
    for c in track(conversations):
        matched = False
        matched_messages = []

        if search in c.data["name"]:
            matched = True
            if preview_count < max_preview:
                matched_messages.append(c.data["name"])

        if search in c.data["prompt"]:
            matched = True
            if preview_count < max_preview:
                matched_messages.append(c.data["prompt"])

        for m in c.data["messages"]:
            if search in m["content"]:
                matched = True
                if preview_count < max_preview:
                    matched_messages.append(m["content"])

        if matched:
            preview_count += 1

            if preview_count < max_preview:
                print(f"Search Result-{preview_count}".center(100, "-"))
                print("[bold red]Original Data[/bold red]")
                print(matched_messages)
                print("[bold green]Replaced Data[/bold green]")
                modified_messages = [
                    _.replace(search, replace) for _ in matched_messages
                ]
                print(modified_messages)

            matched_conversations.append(c)

    logger.info(
        f"Total conversations {db.count_conversations()}, contains [{search}]: {len(matched_conversations)}"
    )

    if run:
        for it in track(matched_conversations, description="replacing string"):
            it.data["name"] = it.data["name"].replace(search, replace)
            it.data["prompt"] = it.data["prompt"].replace(search, replace)
            for m in it.data["messages"]:
                m["content"] = m["content"].replace(search, replace)
            it.updated_at = datetime.utcnow()
            db.update_conversation(it)
        db.vacuum()
    else:
        interactive_view_conversations(db, matched_conversations)



import math
from pathlib import Path

from loguru import logger
import typer
from rich.progress import track

from llm_labeling_ui.db_schema import DBManager

app = typer.Typer(
    add_completion=False,
    pretty_exceptions_show_locals=False,
    short_help="Add tags to you data, such as lang classification(en,zh..), traditional or simplified chinese classification, etc.",
)


@app.command(help="Language Classification")
def lang(
    db_path: Path = typer.Option(..., exists=True, dir_okay=False),
):
    from llm_labeling_ui.lang_classification import LanguageClassifier

    lang_classifier = LanguageClassifier()
    db = DBManager(db_path)
    conversions_count = db.count_conversations()
    logger.info(f"Total conversations: {conversions_count}")
    page_size = 256
    total_pages = math.ceil(conversions_count / page_size)
    for convs in track(db.gen_conversations(batch_size=page_size), total=total_pages):
        for conv in convs:
            tags = conv.data.get("tags", {})
            if tags.get("lang"):
                continue
            lang = lang_classifier(conv.merged_text())
            conv.data["tags"] = {**tags, "lang": lang}
        db.bucket_update_conversation([it.dict() for it in convs])


@app.command(help="Traditional or Simplified Chinese Classification")
def is_traditional_zh(
    db_path: Path = typer.Option(..., exists=True, dir_okay=False),
):
    import opencc

    converter = opencc.OpenCC("t2s.json")

    def is_traditional_chinese(text):
        converted_text = converter.convert(text)
        if converted_text != text:
            return True
        return False

    db = DBManager(db_path)
    conversions_count = db.count_conversations()
    logger.info(f"Total conversations: {conversions_count}")
    page_size = 256
    total_pages = math.ceil(conversions_count / page_size)

    is_traditional_count = 0
    for convs in track(db.gen_conversations(batch_size=page_size), total=total_pages):
        for conv in convs:
            tags = conv.data.get("tags", {})
            is_traditional = is_traditional_chinese(conv.merged_text(role="user"))
            if is_traditional:
                is_traditional_count += 1
            conv.data["tags"] = {
                **tags,
                "is_traditional_zh": is_traditional,
            }
        db.bucket_update_conversation([it.dict() for it in convs])
    logger.info(f"Total traditional chinese conversations: {is_traditional_count}")



from datetime import datetime
from pathlib import Path

import typer
from loguru import logger
from rich.progress import track
from typer import Typer

from llm_labeling_ui.cluster_cmd import app as cluster_app
from llm_labeling_ui.conversation_cmd import app as conversation_app
from llm_labeling_ui.server_cmd import app as server_app
from llm_labeling_ui.tag_cmd import app as tag_app
from llm_labeling_ui.db_schema import DBManager
from llm_labeling_ui.utils import parse_tag

typer_app = Typer(add_completion=False, pretty_exceptions_show_locals=False)
typer_app.add_typer(cluster_app, name="cluster")
typer_app.add_typer(conversation_app, name="conversation")
typer_app.add_typer(tag_app, name="tag")
typer_app.add_typer(server_app, name="server")


@typer_app.command(help="Create db from chatbot-ui history file")
def create_db(
    json_path: Path = typer.Option(
        ..., exists=True, dir_okay=False, help="chatbotui json history file"
    ),
    save_path: Path = typer.Option(None, dir_okay=False),
    force: bool = typer.Option(False, help="force overwrite save_path if exists"),
):
    if save_path is None:
        save_path = json_path.with_suffix(".sqlite")
    logger.info(f"create db at {save_path}")
    if save_path.exists():
        if not force:
            raise FileExistsError(f"{save_path} exists, use --force to overwrite")

    db = DBManager(save_path)
    db.create_from_json_file(json_path)


@typer_app.command(help="Export db to chatbot-ui history file")
def export(
    db_path: Path = typer.Option(None, exists=True, dir_okay=False),
    save_path: Path = typer.Option(
        None,
        dir_okay=False,
        help="If not specified, it will be generated in the same directory as db_path, and the file name will be added with a timestamp.",
    ),
    tag: str = typer.Option(
        "", help="tag to filter conversations. key1,value1,key2,value2..."
    ),
    min_messages: int = typer.Option(0, help="min messages count. included"),
    max_messages: int = typer.Option(10000, help="max messages count. excluded"),
    count: int = typer.Option(-1, help="max conversations to export. -1 for all."),
    force: bool = typer.Option(False, help="force overwrite save_path if exists"),
):
    tags = parse_tag(tag)

    if save_path and save_path.exists():
        if not force:
            raise FileExistsError(f"{save_path} exists, use --force to overwrite")

    if save_path is None:
        save_path = (
            db_path.parent / f"{db_path.stem}_{datetime.utcnow().timestamp()}.json"
        )
    logger.info(f"Dumping db to {save_path}")
    db = DBManager(db_path)
    db.export_to_json_file(
        save_path,
        count,
        min_messages=min_messages,
        max_messages=max_messages,
        tags=tags,
    )


if __name__ == "__main__":
    typer_app()



import distutils
import random
import typing
from typing import Any, Dict, List, Union

from rich.console import Console
from rich.markdown import Markdown
from rich.prompt import Prompt

if typing.TYPE_CHECKING:
    from llm_labeling_ui.db_schema import Conversation

MESSAGE_FILTER_NONE = "message-count-none"
MESSAGE_FILTER_EQUAL = "message-count-equal"
MESSAGE_FILTER_GREATER = "message-count-greater"
MESSAGE_FILTER_LESS = "message-count-less"


def interactive_view_conversations(
    db, id_groups: Union[List[List[str]], List["Conversation"]], max_messages=-1
):
    if len(id_groups) == 0:
        return
    if not isinstance(id_groups[0], list):
        id_groups = [[str(c.id)] for c in id_groups]
    index = 0

    console = Console()
    while True:
        group = id_groups[index]
        conversations = db.get_conversations_by_ids(group)

        markdown_str = ""
        for conv_i, conv in enumerate(conversations):
            markdown_str += f"# {index}-{conv_i}/{len(id_groups)} message count {len(conv.data['messages'])}\n\n"
            if max_messages == -1:
                max_messages = len(conv.data["messages"])
            for m in conv.data["messages"][0:max_messages]:
                markdown_str += f"## {m['role']}\n\n{m['content']}\n\n"

        md = Markdown(markdown_str)
        console.print(md)

        choice = Prompt.ask(
            f"{index}/{len(id_groups)}",
            choices=["h", "l", "r", "q"],
            default="l",
        )
        if choice == "h":
            index -= 1
            if index < 0:
                index = len(id_groups) - 1
        elif choice == "l":
            index += 1
            if index >= len(id_groups):
                index = 0
        elif choice == "r":
            random_index = random.randint(0, len(id_groups) - 1)
            index = random_index
        elif choice == "q":
            break


def str_to_bool(text) -> bool:
    try:
        return bool(distutils.util.strtobool(text))
    except:
        return text


def parse_tag(text: str) -> Dict[str, Any]:
    if not text:
        return {}

    splits = text.split(",")
    if len(splits) % 2 != 0:
        raise ValueError(f"Invalid tag {text}, please use key1,value1,key2,value2,...")

    tags = {}
    for i in range(0, len(splits), 2):
        tags[splits[i]] = str_to_bool(splits[i + 1])
    return tags



from enum import Enum
import json
import math
import random
from typing import List
from more_itertools import flatten

import typer
from pathlib import Path

from loguru import logger
from rich.progress import track
from rich.prompt import Confirm
from rich import print
from rich.markdown import Markdown

from llm_labeling_ui.db_schema import DBManager
from llm_labeling_ui.utils import interactive_view_conversations

app = typer.Typer(
    add_completion=False,
    pretty_exceptions_show_locals=False,
    short_help="Cluster operations, such as create embedding, run cluster, dedup, etc.",
)


@app.command(help="Create embedding")
def create_embedding(
    db_path: Path = typer.Option(..., exists=True, dir_okay=False),
    save_path: Path = typer.Option(
        None,
        dir_okay=False,
        help="Parquet file with column name: id, embedding. If None, embedding will be saved in the same directory as db_path, with parquet file suffix.",
    ),
    model_id: str = typer.Option(
        "BAAI/bge-large-zh-v1.5",
        help="Embedding model id on huggingface. https://huggingface.co/models?pipeline_tag=feature-extraction",
    ),
    max_messages: int = typer.Option(
        1, help="Number of messages used to create embedding. -1 means all messages."
    ),
    device: str = typer.Option("cpu"),
):
    if save_path is None:
        save_path = db_path.with_suffix(".parquet")

    from llm_labeling_ui.cluster import EmbeddingModel
    import pandas as pd

    db = DBManager(db_path)
    model = EmbeddingModel(model_id, device)

    if save_path.exists():
        exists_df = pd.read_parquet(save_path)
        logger.info(f"Load exists embedding: {len(exists_df)}")
    else:
        exists_df = None

    res = []
    total = db.count_conversations()
    logger.info(f"Total conversations: {total}")
    page_size = 256
    for convs in track(
        db.gen_conversations(page_size), total=math.ceil(total / page_size)
    ):
        for conv in convs:
            if exists_df is not None:
                if str(conv.id) in exists_df["id"].values:
                    continue

            messages = []
            for msg in conv.data["messages"]:
                if msg["role"] == "user":
                    messages.append(msg["content"])
            if max_messages == -1:
                max_messages = len(messages)
            messages = messages[:max_messages]
            vector = model(messages)
            res.append({"id": str(conv.id), "embedding": vector})

    df = pd.DataFrame(res)
    if exists_df is not None:
        logger.info(f"df ({len(df)})")
        df = pd.concat([exists_df, df], ignore_index=True)
        logger.info(f"Merge result: {len(df)}")

    df.to_parquet(save_path)


@app.command(help="Remove embedding not exists in db")
def prune_embedding(
    embedding: Path = typer.Option(..., exists=True, dir_okay=False),
    db_path: Path = typer.Option(None, dir_okay=False),
    run: bool = typer.Option(False, help="Run the command"),
):
    import pandas as pd

    if db_path is None:
        db_path = embedding.with_suffix(".sqlite")
    logger.info(f"db_path: {db_path}")

    db = DBManager(db_path)
    ids = [str(it.id) for it in db.all_conversations()]

    df = pd.read_parquet(embedding)
    original_df_len = len(df)
    df = df[df.id.isin(ids)]
    logger.info(f"Original embedding count: {original_df_len}, after prune: {len(df)}")
    if run:
        df.to_parquet(embedding)


class DBSCANMetric(str, Enum):
    euclidean = "euclidean"
    cosine = "cosine"
    manhattan = "manhattan"


@app.command(help="DBSCAN embedding cluster")
def run(
    embedding: Path = typer.Option(
        ...,
        exists=True,
        dir_okay=False,
        help="Parquet file with column name: id, embedding.",
    ),
    save_path: Path = typer.Option(
        None,
        exists=False,
        dir_okay=False,
        help="If None, cluster result will be saved in the same directory as parquet file",
    ),
    force: bool = typer.Option(False, help="Force to run clustering"),
    resume: bool = typer.Option(False, help="Resume cluster result"),
    metric: DBSCANMetric = typer.Option(
        "euclidean", help="DBSCAN metric. euclidean or cosine"
    ),
    eps: float = typer.Option(0.5, help="DBSCAN eps"),
    eps_decay: float = typer.Option(
        0.995, help="DBSCAN eps decay on every epoch. new_eps = eps * eps_decay"
    ),
    min_samples: int = typer.Option(3, help="DBSCAN min_samples"),
    max_samples: int = typer.Option(
        10,
        help="max samples to keep in a cluster",
    ),
    recluster_samples: int = typer.Option(
        80,
        help="If the number of samples in a cluster exceeds recluster_samples, reduce eps cluster again.",
    ),
    epochs: int = typer.Option(5, help="Number of times all data is clustered."),
    bucket_size: int = typer.Option(
        20000, help="The maximum amount of data for a single cluster"
    ),
):
    import pandas as pd
    from llm_labeling_ui.cluster import run_dbscan_cluster

    if save_path is None:
        save_path = embedding.with_suffix(".cluster.json")

    exist_groups = []
    if save_path.exists():
        assert not (
            force and resume
        ), "Cannot use --force and --resume at the same time"

        if force:
            logger.warning(f"Force to run clustering, save result to {save_path}")
        elif resume:
            logger.warning(f"Resume clustering, save result to {save_path}")

            with open(save_path, "r", encoding="utf-8") as fr:
                exist_groups = json.load(fr)["groups"]
        else:
            logger.error(
                f"Cluster result exists: {save_path}, use --force to overwrite or --resume to resume"
            )
            return

    logger.info(f"Save cluster result to {save_path}")

    df = pd.read_parquet(embedding)
    total_samples_count = len(df)
    flatten_exists_groups = set(flatten(exist_groups))
    logger.info(f"Total samples: {total_samples_count}")
    if len(flatten_exists_groups) > 0:
        df = df[~df.id.isin(flatten_exists_groups)]
        logger.info(
            f"Remove {len(flatten_exists_groups)} embedding in exists groups, remain: {len(df)}"
        )

    logger.info(f"Embedding size: {len(df['embedding'].iloc[0])}")
    id_groups = run_dbscan_cluster(
        df,
        metric,
        eps,
        eps_decay,
        min_samples,
        max_samples,
        recluster_samples,
        epochs,
        bucket_size,
    )
    id_groups.extend(exist_groups)
    logger.info(
        f"Total samples: {total_samples_count}, cluster group count: {len(id_groups)}, samples in clusters: {sum([len(it) for it in id_groups])}"
    )
    with open(save_path, "w", encoding="utf-8") as fw:
        json.dump(
            {
                "groups": id_groups,
                "meta": {
                    "eps": eps,
                    "min_samples": min_samples,
                    "max_samples": max_samples,
                    "recluster_samples": recluster_samples,
                    "epochs": epochs,
                    "bucket_size": bucket_size,
                    "total_groups": len(id_groups),
                    "total_samples_in_groups": sum([len(it) for it in id_groups]),
                },
            },
            fw,
            ensure_ascii=False,
            indent=2,
        )


@app.command(help="View cluster result")
def view(
    db_path: Path = typer.Option(..., exists=True, dir_okay=False),
    cluster_path: Path = typer.Option(None, dir_okay=False),
):
    if cluster_path is None:
        cluster_path = db_path.with_suffix(".cluster.json")
        logger.info(f"cluster_path is None, try to load {cluster_path}")
    if not cluster_path.exists():
        logger.error(f"cluster_path not exists: {cluster_path}")
        return

    with open(cluster_path, "r", encoding="utf-8") as fr:
        cluster_result = json.load(fr)
    id_groups: List[List[str]] = cluster_result["groups"]
    id_groups.sort(key=lambda it: len(it), reverse=True)

    db = DBManager(db_path)
    interactive_view_conversations(db, id_groups, max_messages=1)


class DedupStrategy(str, Enum):
    max_messages_count: str = "max_messages_count"
    max_messages_length: str = "max_messages_length"


@app.command(
    help="Delete redundant data in the same clustering result according to certain strategies."
)
def dedup(
    db_path: Path = typer.Option(..., exists=True, dir_okay=False),
    cluster_path: Path = typer.Option(None, dir_okay=False),
    cluster_keep: int = typer.Option(
        1, help="Keep the first N conversations in a cluster"
    ),
    strategy: DedupStrategy = typer.Option(
        DedupStrategy.max_messages_count, help="Dedup strategy"
    ),
    tokenizer: str = typer.Option(
        None, help="Tokenizer name. Used when strategy is max_messages_length"
    ),
    run: bool = typer.Option(False, help="Run delete"),
):
    if strategy == DedupStrategy.max_messages_length:
        assert (
            tokenizer is not None
        ), "tokenizer must be specified when strategy is max_messages_length"
        from transformers import AutoTokenizer

        tokenizer = AutoTokenizer.from_pretrained(tokenizer, trust_remote_code=True)

    if cluster_path is None:
        cluster_path = db_path.with_suffix(".cluster.json")
        logger.info(f"cluster_path is None, try to load {cluster_path}")
    if not cluster_path.exists():
        logger.error(f"cluster_path not exists: {cluster_path}")
        return

    with open(cluster_path, "r", encoding="utf-8") as fr:
        cluster_result = json.load(fr)
    id_groups: List[List[str]] = cluster_result["groups"]
    logger.info(f"Total groups: {len(id_groups)}")

    db = DBManager(db_path)
    total_conversations = db.count_conversations()
    deleted_count = 0
    if not run:
        random.shuffle(id_groups)

    for group in track(id_groups, disable=not run):
        convs = db.get_conversations_by_ids(group)
        if strategy == DedupStrategy.max_messages_count:
            convs.sort(key=lambda it: it.messages_count(), reverse=True)
        elif strategy == DedupStrategy.max_messages_length:
            convs.sort(key=lambda it: it.token_count(), reverse=True)

        convs_to_keep = convs[:cluster_keep]
        convs_to_delete = convs[cluster_keep:]
        if run:
            db.delete_conversation([it.id for it in convs_to_delete])
            deleted_count += len(convs_to_delete)
        else:
            print(Markdown(f"# Conversations to keep: ({len(convs_to_keep)})"))

            for c in convs_to_keep:
                print(Markdown(f"## Message count: {c.messages_count()}. {c.id}"))
                print(c.data["messages"][:2])

            print(Markdown(f"# Conversations to delete: ({len(convs_to_delete)})"))
            for c in convs_to_delete:
                print(Markdown(f"## Message count: {c.messages_count()}. {c.id}"))
                print(c.data["messages"][:2])

            if Confirm.ask("Press Enter to continue", default=True):
                continue
            else:
                exit(0)

    db.vacuum()
    print(
        f"Total conversations: {total_conversations}, delete {deleted_count} conversations, remain {total_conversations - deleted_count} conversations"
    )



import json
from datetime import datetime
import math
from pathlib import Path
import random
from typing import Iterator, Optional, Dict, List, Union
from uuid import UUID, uuid4

import sqlmodel
from loguru import logger
from rich.progress import track
from sqlalchemy import Column, select
from sqlmodel import SQLModel, Field, create_engine, Session, JSON, col
from sqlalchemy import func

from llm_labeling_ui.utils import (
    MESSAGE_FILTER_EQUAL,
    MESSAGE_FILTER_GREATER,
    MESSAGE_FILTER_LESS,
    MESSAGE_FILTER_NONE,
)


class TimestampModel(SQLModel):
    created_at: datetime = Field(default_factory=datetime.utcnow)
    updated_at: Optional[datetime]


class UUIDIDModel(SQLModel):
    id: UUID = Field(default_factory=uuid4, primary_key=True)


class Conversation(UUIDIDModel, TimestampModel, table=True):
    data: Dict = Field(default={}, sa_column=Column(JSON))

    def contain_tags(self, tags: Dict) -> bool:
        flag = True
        for k, v in tags.items():
            if self.data.get("tags", {}).get(k) != v:
                flag = False
                break
        return flag

    def merged_text(self, max_messages: int = -1, role: str = "all") -> str:
        if max_messages == -1:
            max_messages = len(self.data["messages"])

        if role == "all":
            return "".join(
                [self.data["prompt"]]
                + [m["content"] for m in self.data["messages"][0:max_messages]]
            )
        elif role == "system":
            return self.data["prompt"]
        elif role == "user":
            return "".join(
                [
                    m["content"]
                    for m in self.data["messages"][0:max_messages]
                    if m["role"] == "user"
                ]
            )
        elif role == "assistant":
            return "".join(
                [
                    m["content"]
                    for m in self.data["messages"][0:max_messages]
                    if m["role"] == "assistant"
                ]
            )
        else:
            raise ValueError(f"Invalid role {role}")

    def messages_count(self) -> int:
        return len(self.data["messages"])

    def token_count(self, tokenizer) -> int:
        return len(tokenizer(self.merged_text())["input_ids"])

    class Config:
        arbitrary_types_allowed = True


class Folder(UUIDIDModel, TimestampModel, table=True):
    name: str
    type: str = "chat"


class PromptTemp(UUIDIDModel, TimestampModel, table=True):
    name: str
    description: str
    content: str
    model: Dict = Field(default={}, sa_column=Column(JSON))
    folderId: Optional[UUID] = None


class DBManager:
    def __init__(self, db_path: Path):
        self.engine = create_engine(
            f"sqlite:///{db_path}",
            json_serializer=lambda obj: json.dumps(obj, ensure_ascii=False),
        )
        SQLModel.metadata.create_all(self.engine)

    def create_from_json_file(self, json_p: Path) -> "DBManager":
        from llm_labeling_ui.schema import ChatBotUIHistory

        with open(json_p, "r", encoding="utf-8") as f:
            chatbot_ui_history = ChatBotUIHistory.parse_raw(f.read())

        with Session(self.engine) as session:
            for it in track(
                chatbot_ui_history.history, description="writing history to db"
            ):
                session.add(Conversation(id=it.id, data=it.dict()))
            session.commit()

            for it in chatbot_ui_history.folders:
                session.add(Folder(data=it.dict()))
            session.commit()

            for it in chatbot_ui_history.prompts:
                session.add(PromptTemp(data=it.dict()))
            session.commit()
        return self

    def export_to_json_file(
        self,
        json_p: Path,
        count: int,
        min_messages: int,
        max_messages: int,
        tags: Dict = {},
    ):
        from llm_labeling_ui.schema import (
            ChatBotUIHistory,
            Conversation as UIConversation,
        )

        chatbot_ui_history = ChatBotUIHistory()
        chatbot_ui_history.folders = self.get_folders()
        chatbot_ui_history.prompts = self.get_prompt_temps()
        with Session(self.engine) as session:
            statement = sqlmodel.select(Conversation)
            results = session.exec(statement).all()

            if tags:
                _new_results = []
                for it in results:
                    if it.contain_tags(tags):
                        _new_results.append(it)
                results = _new_results

            logger.info(f"Shuffling {len(results)} conversations")
            random.shuffle(results)
            for it in track(results):
                if min_messages <= it.messages_count() < max_messages:
                    chatbot_ui_history.history.append(UIConversation(**it.data))

            if count == -1:
                count = len(results)
            chatbot_ui_history.history = chatbot_ui_history.history[:count]

        logger.info(f"export {len(chatbot_ui_history.history)} conversations")
        with open(json_p, "w", encoding="utf-8") as f:
            json.dump(chatbot_ui_history.dict(), f, ensure_ascii=False)

    def get_folders(self) -> List[Folder]:
        with Session(self.engine) as session:
            statement = sqlmodel.select(Folder)
            folders = session.exec(statement).all()
            return folders

    def get_prompt_temps(self) -> List[PromptTemp]:
        with Session(self.engine) as session:
            statement = sqlmodel.select(PromptTemp)
            prompts = session.exec(statement).all()
            return prompts

    def get_conversations(
        self,
        page: int,
        page_size: int = 50,
        search_term: Union[str, List[str]] = "",
        messageCountFilterCount: int = 0,
        messageCountFilterMode: str = MESSAGE_FILTER_NONE,
    ) -> List[Conversation]:
        limit = page_size
        offset = page * page_size
        with Session(self.engine) as session:
            statement = (
                sqlmodel.select(Conversation)
                .order_by(Conversation.created_at.desc())
                .offset(offset)
                .limit(limit)
            )
            statement = self._filter(
                statement, search_term, messageCountFilterCount, messageCountFilterMode
            )
            convs = session.exec(statement).all()
            return convs

    def get_conversations_by_ids(
        self,
        ids: List[str],
    ) -> List[Conversation]:
        with Session(self.engine) as session:
            statement = sqlmodel.select(Conversation).where(Conversation.id.in_(ids))
            convs = session.exec(statement).all()
            return convs

    def all_conversations(
        self, search_term: Union[str, List[str]] = ""
    ) -> List[Conversation]:
        return self.get_conversations(0, 1000000000, search_term=search_term)

    def gen_conversations(self, batch_size) -> Iterator[List[Conversation]]:
        total = self.count_conversations()
        total_pages = math.ceil(total / batch_size)
        for page in range(total_pages):
            yield self.get_conversations(page, batch_size)

    def count_conversations(
        self,
        search_term: str = "",
        messageCountFilterCount: int = 0,
        messageCountFilterMode: str = MESSAGE_FILTER_NONE,
    ) -> int:
        with Session(self.engine) as session:
            statement = select(func.count(Conversation.id))
            statement = self._filter(
                statement, search_term, messageCountFilterCount, messageCountFilterMode
            )
            convs = session.exec(statement).all()
            return convs[0][0]

    def update_conversation(self, conv: Conversation):
        with Session(self.engine) as session:
            statement = select(Conversation).where(Conversation.id == conv.id)
            exist_conv = session.exec(statement).one()[0]
            exist_conv.data = conv.data
            exist_conv.updated_at = datetime.utcnow()
            session.add(exist_conv)
            session.commit()
            session.refresh(exist_conv)
            # return exist_conv

    def bucket_update_conversation(self, convs: List[Conversation]):
        with Session(self.engine) as session:
            session.bulk_update_mappings(Conversation, convs)
            session.flush()
            session.commit()

    def create_conversation(self, conv: Conversation):
        with Session(self.engine) as session:
            session.add(conv)
            session.commit()
            # return conv

    def delete_conversation(self, id: Union[str, List[str]]):
        if not isinstance(id, list):
            id = [id]

        with Session(self.engine) as session:
            statement = select(Conversation).where(Conversation.id.in_(id))
            results = session.exec(statement).all()
            for r in results:
                session.delete(r[0])
            session.commit()

    def vacuum(self):
        with Session(self.engine) as session:
            session.execute("VACUUM")

    def _filter(
        self, statement, search_term, messageCountFilterCount, messageCountFilterMode
    ):
        if messageCountFilterMode == MESSAGE_FILTER_EQUAL:
            statement = statement.where(
                func.json_array_length(Conversation.data.op("->>")("messages"))
                == messageCountFilterCount
            )
        elif messageCountFilterMode == MESSAGE_FILTER_GREATER:
            statement = statement.where(
                func.json_array_length(Conversation.data.op("->>")("messages"))
                > messageCountFilterCount
            )
        elif messageCountFilterMode == MESSAGE_FILTER_LESS:
            statement = statement.where(
                func.json_array_length(Conversation.data.op("->>")("messages"))
                < messageCountFilterCount
            )

        if search_term:
            if isinstance(search_term, str):
                search_term = [search_term]

            for s in search_term:
                statement = statement.where(col(Conversation.data).contains(s))

        return statement



import math
from typing import List
import pandas as pd
import numpy as np
from loguru import logger
from sklearn.cluster import DBSCAN, OPTICS
from more_itertools import flatten
from transformers import AutoTokenizer, AutoModel
import torch


def _inter_run_dbscan_cluster(
    df: pd.DataFrame,
    metric: str,
    eps: float,
    min_samples: int,
    max_samples: int,
    recluster_samples: int,
    bucket_size: int,
) -> List[List[str]]:
    """
    Run DBSCAN clustering on a dataframe's embedding column.

    Args:
        df:
        eps: DBSCAN eps. The maximum distance between two samples for one to be considered
            as in the neighborhood of the other.
        min_samples:
        max_samples: If the number of samples in a cluster exceeds max_samples, multiply eps by 2/3 and cluster again.
        bucket_size: The maximum amount of data for a single cluster

    Returns: The returned clustering results, the data of each group is between [min_samples, max_samples].
    """

    def predict(X, eps) -> List[List[int]]:
        # metric = "euclidean"
        # metric = "cosine"
        clusterer = DBSCAN(eps=eps, min_samples=min_samples, metric=metric)
        # clusterer = OPTICS(
        #     min_samples=min_samples,
        #     max_eps=eps,
        #     metric=metric,
        #     cluster_method="dbscan",
        # ).fit(X.tolist())
        y_db = clusterer.fit_predict(X.tolist())

        unique_labels = np.unique(y_db)
        res = []
        for label in unique_labels:
            if label == -1:
                continue
            indexes = np.where(y_db == label)[0]
            res.append(indexes.tolist())
        return res

    total_buckets = math.ceil(len(df) / bucket_size)
    all_id_groups: List[List[str]] = []
    for i, bucket in enumerate(np.array_split(df, total_buckets)):
        logger.info(f"Processing bucket {i + 1}/{total_buckets}, size: {len(bucket)}")
        index_groups = predict(bucket["embedding"], eps)
        logger.info(
            f"bucket groups: {len(index_groups)}. {sorted([len(it) for it in index_groups], reverse=True)}"
        )
        for index_group in index_groups:
            id_group = bucket.iloc[index_group].id.tolist()
            if len(id_group) <= max_samples:
                all_id_groups.append(id_group)
            elif len(id_group) >= recluster_samples:
                large_sub_bucket = bucket[bucket["id"].isin(id_group)]
                sub_index_groups = predict(
                    large_sub_bucket["embedding"], eps=eps * 4 / 5
                )
                sub_id_groups = [
                    large_sub_bucket.iloc[sub_index_group].id.tolist()
                    for sub_index_group in sub_index_groups
                ]
                if len(sub_id_groups):
                    logger.info(
                        f"Group size: {len(large_sub_bucket)} > recluster_samples({recluster_samples}), recluster {len(sub_id_groups)} sub groups: {[len(it) for it in sub_id_groups]}"
                    )
                for it in sub_id_groups:
                    if len(it) <= max_samples:
                        all_id_groups.append(it)
    return all_id_groups


def run_dbscan_cluster(
    df: pd.DataFrame,
    metric: str,
    eps: float,
    eps_decay: float,
    min_samples: int,
    max_samples: int,
    recluster_samples: int,
    epochs: int,
    bucket_size: int,
) -> List[List[str]]:
    """
    Run DBSCAN clustering on a dataframe's embedding column.

    Args:
        df:
        eps: DBSCAN eps. The maximum distance between two samples for one to be considered
            as in the neighborhood of the other.
        min_samples:
        max_samples: If the number of samples in a cluster exceeds max_samples, multiply eps by 0.5 and cluster again.
        epochs: Number of times all data is clustered.
        bucket_size: The maximum amount of data for a single cluster

    Returns: The returned clustering results, the data of each group is between [min_samples, max_samples].
    """
    all_id_groups: List[List[str]] = []
    for iter in range(epochs):
        total_samples_count = len(df)
        logger.info(
            f"Running DBSCAN clustering epoch: {iter + 1}/{epochs}, total samples: {total_samples_count}"
        )
        id_groups = _inter_run_dbscan_cluster(
            df, metric, eps, min_samples, max_samples, recluster_samples, bucket_size
        )
        new_eps = eps * eps_decay
        logger.info(f"Decay eps: {eps} -> {new_eps}")
        eps = new_eps
        clustered_ids = set(flatten(id_groups))
        unclustered_ids = set(df.id.tolist()) - clustered_ids
        df = df[df.id.isin(unclustered_ids)]
        df = df.sample(frac=1)
        all_id_groups.extend(id_groups)
        if len(df) == total_samples_count:
            logger.warning(f"DBSCAN clustering early stop at epoch: {iter + 1}")
            break

    return all_id_groups


# Sentences we want sentence embeddings for
class EmbeddingModel:
    def __init__(self, model_id, device):
        # Load model from HuggingFace Hub
        self.tokenizer = AutoTokenizer.from_pretrained(model_id)
        self.model = AutoModel.from_pretrained(model_id).eval().to(device)
        self.device = device

    @torch.inference_mode()
    def __call__(self, texts: List[str]) -> List[float]:
        # Tokenize sentences
        encoded_input = self.tokenizer(
            texts, padding=True, truncation=True, return_tensors="pt", max_length=512
        )
        # for s2p(short query to long passage) retrieval task, add an instruction to query (not add instruction for passages)
        # encoded_input = tokenizer([instruction + q for q in queries], padding=True, truncation=True, return_tensors='pt')

        encoded_input = {k: v.to(self.device) for k, v in encoded_input.items()}

        # Compute token embeddings
        model_output = self.model(**encoded_input)
        # Perform pooling. In this case, cls pooling.
        sentence_embeddings = model_output[0].mean(dim=0)[0]
        # normalize embeddings
        sentence_embeddings = torch.nn.functional.normalize(
            sentence_embeddings, p=2, dim=0
        )
        return sentence_embeddings.cpu().tolist()



import math
import os
from functools import lru_cache
from typing import List
from pathlib import Path
from fastapi import APIRouter, FastAPI, HTTPException
from fastapi.responses import FileResponse, StreamingResponse
from fastapi.staticfiles import StaticFiles
import openai
from loguru import logger

from llm_labeling_ui.db_schema import (
    DBManager,
    Folder,
    PromptTemp,
    Conversation as DBConversation,
)
from llm_labeling_ui.schema import (
    ChatMessage,
    ChatRequest,
    ModelsRequest,
    OpenAIModelID,
    Config,
    GetConversionsRequest,
    GetConversionsResponse,
    Conversation,
    CountTokensResponse,
    CountTokensRequest,
)

error503 = "OpenAI server is busy, try again later"


class Api:
    def __init__(self, app: FastAPI, config: Config, db: DBManager, tokenizer=None):
        self.router = APIRouter()
        self.app = app
        self.db = db
        self.tokenizer = tokenizer
        self.config = config
        if self.tokenizer is not None:
            from transformers import AutoTokenizer

            self.tokenizer = AutoTokenizer.from_pretrained(
                self.tokenizer, trust_remote_code=True
            )

        self.app.mount(
            "/static", StaticFiles(directory=config.web_app_dir), name="static"
        )

        self.add_api_route(
            "/",
            self.main,
            methods=["GET"],
        )

        self.add_api_route(
            "/api/models",
            self.models,
            methods=["POST"],
            response_model=List[dict],
        )

        self.add_api_route(
            "/api/chat",
            self.chat,
            methods=["POST"],
            response_model=str,
        )

        self.add_api_route(
            "/api/folders",
            self.get_folders,
            methods=["GET"],
            response_model=List[Folder],
        )

        self.add_api_route(
            "/api/prompt_temps",
            self.get_prompt_temps,
            methods=["GET"],
            response_model=List[PromptTemp],
        )

        self.add_api_route(
            "/api/conversations",
            self.get_conversations,
            methods=["POST"],
            response_model=GetConversionsResponse,
        )

        self.add_api_route(
            "/api/create_conversation",
            self.create_conversation,
            methods=["POST"],
            # response_model=Conversation,
        )

        self.add_api_route(
            "/api/update_conversation",
            self.update_conversation,
            methods=["POST"],
            # response_model=Conversation,
        )

        self.add_api_route(
            "/api/delete_conversation",
            self.delete_conversation,
            methods=["POST"],
            # response_model=Conversation,
        )

        self.add_api_route(
            "/api/count_tokens",
            self.count_tokens,
            methods=["POST"],
            response_model=CountTokensResponse,
        )

    def main(self):
        return FileResponse(os.path.join(self.config.web_app_dir, "index.html"))

    def models(self, req: ModelsRequest) -> List[dict]:
        api_key = req.key if req.key else os.environ.get("OPENAI_API_KEY")
        api_org = req.org if req.org else os.environ.get("OPENAI_ORGANIZATION")
        all_models = openai.Model.list(api_key, organization=api_org)
        res = []
        for it in all_models.data:
            if it.id in OpenAIModelID:
                res.append(
                    {
                        "id": it.id,
                        "name": OpenAIModelID[it.id],
                    }
                )
        return res

    def chat(self, req: ChatRequest):
        messages = req.messages
        if req.prompt:
            messages.insert(0, ChatMessage(role="system", content=req.prompt))

        if req.key:
            openai.api_key = req.key
        else:
            openai.api_key = os.environ.get("OPENAI_API_KEY")

        if req.org:
            openai.organization = req.org
        else:
            openai.organization = os.environ.get("OPENAI_ORGANIZATION")

        def gen():
            try:
                stream_response = openai.ChatCompletion.create(
                    model=req.model.id,
                    messages=[it.dict() for it in messages],
                    max_tokens=1000,
                    temperature=req.temperature,
                    stream=True,
                )
            except Exception as e:
                logger.error("Error in creating campaigns from openAI:", str(e))
                raise HTTPException(503, error503)

            try:
                for chunk in stream_response:
                    current_content = chunk["choices"][0]["delta"].get("content", "")
                    yield current_content

            except Exception as e:
                logger.error("OpenAI Response (Streaming) Error: " + str(e))
                raise HTTPException(503, error503)

        return StreamingResponse(content=gen(), media_type="text/event-stream")

    def get_folders(self) -> List[Folder]:
        return self.db.get_folders()

    def get_prompt_temps(self) -> List[PromptTemp]:
        return self.db.get_prompt_temps()

    def get_conversations(self, req: GetConversionsRequest) -> GetConversionsResponse:
        conversions_count = self.db.count_conversations(
            req.searchTerm, req.messageCountFilterCount, req.messageCountFilterMode
        )
        total_pages = math.ceil(conversions_count / req.pageSize)
        return GetConversionsResponse(
            conversations=self.db.get_conversations(
                page=req.page,
                page_size=req.pageSize,
                search_term=req.searchTerm,
                messageCountFilterCount=req.messageCountFilterCount,
                messageCountFilterMode=req.messageCountFilterMode,
            ),
            page=req.page,
            totalPages=total_pages,
            totalConversations=conversions_count,
        )

    def update_conversation(self, req: Conversation):
        db_req = DBConversation(id=req.id, data=req.dict())
        self.db.update_conversation(db_req)
        return "ok", 200
        # return Conversation(**db_res.data)

    def create_conversation(self, req: Conversation):
        db_req = DBConversation(id=req.id, data=req.dict())
        self.db.create_conversation(db_req)
        return "ok", 200
        # return Conversation(**db_res.data)

    def delete_conversation(self, req: Conversation):
        self.db.delete_conversation(req.id)
        return "ok", 200

    def count_tokens(self, req: CountTokensRequest) -> CountTokensResponse:
        @lru_cache(maxsize=512)
        def cached_count_tokens(text: str) -> int:
            if self.tokenizer is None:
                return 0
            else:
                return len(self.tokenizer(text)["input_ids"])

        count = cached_count_tokens(req.text)
        return CountTokensResponse(count=count)

    def add_api_route(self, path: str, endpoint, **kwargs):
        return self.app.add_api_route(path, endpoint, **kwargs)



import uuid
from typing import List
from pathlib import Path

from pydantic import BaseModel, Field
from sqlmodel import SQLModel
from llm_labeling_ui.utils import MESSAGE_FILTER_NONE
from llm_labeling_ui.db_schema import Conversation as DBConversation

OpenAIModelID = {
    "gpt-3.5-turbo": "GPT_3_5",
    "gpt-35-turbo": "GPT_3_5_AZ",
    "gpt-4": "GPT_4",
    "gpt-4-32k": "GPT_4_32K",
}


class ModelsRequest(BaseModel):
    key: str
    org: str


class ChatMessage(BaseModel):
    role: str
    content: str


class OpenAIModel(BaseModel):
    id: str
    name: str
    maxLength: int
    tokenLimit: int


class ChatRequest(BaseModel):
    model: OpenAIModel
    messages: List[ChatMessage]
    key: str
    org: str
    prompt: str = ""
    temperature: float = 1.0


class ModelInfo(SQLModel):
    id: str = "gpt-3.5-turbo"
    name: str = "GPT-3.5"
    maxLength: int = 12000
    tokenLimit: int = 4000


class Folder(BaseModel):
    id: str = Field(default_factory=uuid.uuid4)
    name: str
    type: str = "chat"


class PromptTemp(BaseModel):
    id: str = Field(default_factory=uuid.uuid4)
    name: str
    description: str
    content: str
    model: ModelInfo
    folderId: str = None


class Message(BaseModel):
    role: str
    content: str


class Conversation(BaseModel):
    id: str = Field(default_factory=uuid.uuid4)
    name: str = ""
    messages: List[Message] = []
    model: ModelInfo = ModelInfo()
    # system
    prompt: str = ""
    temperature: int = 1
    folderId: str = None

    def merged_text(self) -> str:
        return "".join([self.prompt] + [m.content for m in self.messages])


class ChatBotUIHistory(BaseModel):
    version: int = 4
    history: List[Conversation] = []
    folders: List[Folder] = []
    prompts: List[PromptTemp] = []


class Config(BaseModel):
    web_app_dir: Path


class GetConversionsRequest(BaseModel):
    page: int = 0
    pageSize: int = 50
    searchTerm: str = ""
    messageCountFilterCount: int = 0
    messageCountFilterMode: str = MESSAGE_FILTER_NONE


class GetConversionsResponse(BaseModel):
    page: int
    totalPages: int
    conversations: List[DBConversation]
    totalConversations: int


class CountTokensRequest(BaseModel):
    text: str


class CountTokensResponse(BaseModel):
    count: int



import os

import setuptools
from pathlib import Path

web_files = Path("llm_labeling_ui/out/").glob("**/*")
web_files = [str(it).replace("llm_labeling_ui/", "") for it in web_files]

CURRENT_DIR = Path(os.path.abspath(os.path.dirname(__file__)))
with open(CURRENT_DIR / "README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def load_requirements():
    requirements_file_name = "requirements.txt"
    requires = []
    with open(requirements_file_name) as f:
        for line in f:
            if line:
                requires.append(line.strip())
    return requires


setuptools.setup(
    name="llm-labeling-ui",
    version="0.8.0",
    author="PanicByte",
    author_email="cwq1913@gmail.com",
    description="LLM Labeling UI is an open source project for large language model data labeling",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Sanster/llm-labeling-ui",
    packages=setuptools.find_packages("./"),
    package_data={"llm_labeling_ui": web_files},
    install_requires=load_requirements(),
    python_requires=">=3.7",
    entry_points={"console_scripts": ["llm-labeling-ui=llm_labeling_ui:entry_point"]},
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import os
from pathlib import Path

import typer
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from gunicorn.app.base import BaseApplication
from loguru import logger
import typer

from llm_labeling_ui.db_schema import DBManager
from llm_labeling_ui.schema import Config

app = typer.Typer(
    add_completion=False,
    pretty_exceptions_show_locals=False,
    short_help="Start the web server",
)

CURRENT_DIR = Path(os.path.abspath(os.path.dirname(__file__)))
web_app_dir = CURRENT_DIR / "out"


class StandaloneApplication(BaseApplication):
    def __init__(self, app, options, config, db, tokenizer):
        self.options = options or {}
        self.app = app
        self.config = config
        self.db = db
        self.tokenizer = tokenizer
        super().__init__()

    def load_config(self):
        config = {
            key: value
            for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.app


def app_factory():
    app = FastAPI()
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    return app


def post_worker_init(worker):
    from llm_labeling_ui.api import Api

    api = Api(worker.app.app, worker.app.config, worker.app.db, worker.app.tokenizer)
    api.app.include_router(api.router)


@app.command(help="Start the web server")
def start(
    host: str = typer.Option("0.0.0.0"),
    port: int = typer.Option(8000),
    data: Path = typer.Option(
        ..., exists=True, dir_okay=False, help="json or sqlite file"
    ),
    tokenizer: str = typer.Option(None),
):
    config = Config(web_app_dir=web_app_dir)
    options = {
        "bind": f"{host}:{port}",
        # 'workers': workers,
        "worker_class": "uvicorn.workers.UvicornWorker",
        "timeout": 120,
        "post_worker_init": post_worker_init,
        "capture_output": True,
    }

    if data.suffix == ".json":
        db_path = data.with_suffix(".sqlite")
    elif data.suffix == ".sqlite":
        db_path = data
    else:
        raise ValueError(f"unknown file type {data}")

    if not db_path.exists():
        logger.info(f"create db at {db_path}")
        db = DBManager(db_path)
        db = db.create_from_json_file(data)
    else:
        logger.warning(f"loading db from {db_path}, data may be different from {data}")
        db = DBManager(db_path)

    StandaloneApplication(app_factory(), options, config, db, tokenizer).run()
